import dataclasses
import io
import warnings
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd
import polars as pl
from typing_extensions import Self

from neuralib.atlas.data import load_bg_structure_tree
from neuralib.io.core import CCF_CACHE_DIRECTORY, ATLAS_CACHE_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.verbose import fprint, print_save


def _cache_ndarray(url: str, file: PathLike) -> None:
    fprint(f'DOWNLOADING... {file.name} from {url}', vtype='io')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        content = download_with_tqdm(url)
        data = np.load(content, allow_pickle=True)
        np.save(file, data)


def _load_ccf_annotation(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Annotation volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return: Array[uint16, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'annotation_volume_10um_by_index.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925493'
        _cache_ndarray(url, file)

    return np.load(file, allow_pickle=True)


def _load_ccf_template(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Template volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return: Array[uint16, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'template_volume_10um.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925496'
        _cache_ndarray(url, file)

    return np.load(file, allow_pickle=True)


def _load_structure_tree(version: Literal['2017', 'old'] = '2017',
                         output_dir: PathLike | None = None) -> pl.DataFrame:
    """
    Load structure tree dataframe

    :param version: {'2017', 'old'}
    :param output_dir: output directory for caching
    :return: structure tree dataframe

    ::

        ┌───────────┬──────────┬───────────────────────────────┬─────────┬──────────┬─────────────┬───────────────┬────────┬─────────────────────┬───────┬──────────┬─────────────┬───────────────────┬───────────────────┬─────────────────────────┬──────────────────────────────┬────────┬───────────┬──────────────────────┬──────────────┬───────────────────────────────┐
        │ id        ┆ atlas_id ┆ name                          ┆ acronym ┆ st_level ┆ ontology_id ┆ hemisphere_id ┆ weight ┆ parent_structure_id ┆ depth ┆ graph_id ┆ graph_order ┆ structure_id_path ┆ color_hex_triplet ┆ neuro_name_structure_id ┆ neuro_name_structure_id_path ┆ failed ┆ sphinx_id ┆ structure_name_facet ┆ failed_facet ┆ safe_name                     │
        │ ---       ┆ ---      ┆ ---                           ┆ ---     ┆ ---      ┆ ---         ┆ ---           ┆ ---    ┆ ---                 ┆ ---   ┆ ---      ┆ ---         ┆ ---               ┆ ---               ┆ ---                     ┆ ---                          ┆ ---    ┆ ---       ┆ ---                  ┆ ---          ┆ ---                           │
        │ i64       ┆ f64      ┆ str                           ┆ str     ┆ str      ┆ i64         ┆ i64           ┆ i64    ┆ f64                 ┆ i64   ┆ i64      ┆ i64         ┆ list[i64]         ┆ str               ┆ str                     ┆ str                          ┆ str    ┆ i64       ┆ i64                  ┆ i64          ┆ str                           │
        ╞═══════════╪══════════╪═══════════════════════════════╪═════════╪══════════╪═════════════╪═══════════════╪════════╪═════════════════════╪═══════╪══════════╪═════════════╪═══════════════════╪═══════════════════╪═════════════════════════╪══════════════════════════════╪════════╪═══════════╪══════════════════════╪══════════════╪═══════════════════════════════╡
        │ 997       ┆ -1.0     ┆ root                          ┆ root    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ -1.0                ┆ 0     ┆ 1        ┆ 0           ┆ [997]             ┆ FFFFFF            ┆ null                    ┆ null                         ┆ f      ┆ 1         ┆ 385153371            ┆ 734881840    ┆ root                          │
        │ 8         ┆ 0.0      ┆ Basic cell groups and regions ┆ grey    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1           ┆ [997, 8]          ┆ BFDAE3            ┆ null                    ┆ null                         ┆ f      ┆ 2         ┆ 2244697386           ┆ 734881840    ┆ Basic cell groups and regions │
        │ …         ┆ …        ┆ …                             ┆ …       ┆ …        ┆ …           ┆ …             ┆ …      ┆ …                   ┆ …     ┆ …        ┆ …           ┆ …                 ┆ …                 ┆ …                       ┆ …                            ┆ …      ┆ …         ┆ …                    ┆ …            ┆ …                             │
        │ 65        ┆ 715.0    ┆ parafloccular sulcus          ┆ pfs     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1040.0              ┆ 3     ┆ 1        ┆ 1324        ┆ [997, 1024, … 65] ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1325      ┆ 771629690            ┆ 734881840    ┆ parafloccular sulcus          │
        │ 624       ┆ 926.0    ┆ Interpeduncular fossa         ┆ IPF     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1024.0              ┆ 2     ┆ 1        ┆ 1325        ┆ [997, 1024, 624]  ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1326      ┆ 1476705011           ┆ 734881840    ┆ Interpeduncular fossa         │
        │ 304325711 ┆ -2.0     ┆ retina                        ┆ retina  ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1326        ┆ [997, 304325711]  ┆ 7F2E7E            ┆ null                    ┆ null                         ┆ f      ┆ 1327      ┆ 3295290839           ┆ 734881840    ┆ retina                        │
        └───────────┴──────────┴───────────────────────────────┴─────────┴──────────┴─────────────┴───────────────┴────────┴─────────────────────┴───────┴──────────┴─────────────┴───────────────────┴───────────────────┴─────────────────────────┴──────────────────────────────┴────────┴───────────┴──────────────────────┴──────────────┴───────────────────────────────┘

    """
    #
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)
    #
    if version == '2017':
        filename = 'structure_tree_safe_2017.csv'
    elif version == 'old':
        filename = 'structure_tree_safe.csv'
    else:
        raise ValueError('')

    #
    output = output_dir / filename
    if not output.exists():
        url = f'https://raw.githubusercontent.com/cortex-lab/allenCCF/master/{filename}'
        content = download_with_tqdm(url)
        df = pd.read_csv(content)
        df.to_csv(output)
        print_save(output, verb='DOWNLOAD')

    ret = (pl.read_csv(output)
           .with_columns(pl.col('parent_structure_id').fill_null(-1))
           .with_columns(pl.col('structure_id_path')
                         .map_elements(lambda it: tuple(map(int, it[1:-1].split('/'))),
                                       return_dtype=pl.List(pl.Int64))))
    #
    if '2017' in Path(output).name:
        ret = ret.with_columns(pl.col('atlas_id').fill_null(-2))
    else:
        ret = ret.with_columns(pl.col('name').alias('safe_name'))

    return ret


# ========= #


@dataclasses.dataclass
class _CellAtlas:
    dataframe: pl.DataFrame
    """
    Example::

        ┌────────────────────────────────┬────────────────┬───────────┐
        │ Brain region                   ┆ Volumes [mm^3] ┆ n_neurons │
        │ ---                            ┆ ---            ┆ ---       │
        │ str                            ┆ f64            ┆ i64       │
        ╞════════════════════════════════╪════════════════╪═══════════╡
        │ Abducens nucleus               ┆ 0.015281       ┆ 1324      │
        │ Accessory facial motor nucleus ┆ 0.013453       ┆ 497       │
        │ Accessory olfactory bulb       ┆ 0.6880625      ┆ 189608    │
        │ …                              ┆ …              ┆ …         │
        │ Zona incerta                   ┆ 2.157641       ┆ 136765    │
        │ posteromedial visual area      ┆ 1.2225625      ┆ 197643    │
        └────────────────────────────────┴────────────────┴───────────┘
    """

    @classmethod
    def load_from_csv(cls, file: Path | None = None,
                      ignore_cell_types_info: bool = True,
                      ignore_detail_info: bool = True) -> Self:
        """
        Load/Download the csv file

        :param file: filepath. If None, download from source paper
        :param ignore_cell_types_info: ignore cell types information, only select neuron and volume foreach areas
        :param ignore_detail_info: ignore information in brain subregion
        :return:
        """
        if file is None:
            d = ATLAS_CACHE_DIRECTORY

            if not d.exists():
                d.mkdir(exist_ok=True, parents=True)

            file = d / 'cellatlas.csv'
            if not file.exists():
                cls._request(file)

        df = pl.read_csv(file)

        if ignore_cell_types_info:
            df = df.select('Brain region', 'Neuron [mm^-3]', 'Volumes [mm^3]')

        # total neurons
        df = (
            df.with_columns((pl.col('Neuron [mm^-3]') * pl.col('Volumes [mm^3]')).alias('n_neurons').cast(pl.Int64))
            .drop('Neuron [mm^-3]')
        )

        if ignore_detail_info:
            patterns = (',', '/', r'\(')
            for pt in patterns:
                df = df.filter(~(pl.col('Brain region').str.contains(pt)))

        return cls(df.sort('Brain region'))

    @classmethod
    def _request(cls, output: Path) -> pl.DataFrame:
        """download from paper source"""
        import requests

        url = 'https://journals.plos.org/ploscompbiol/article/file?type=supplementary&id=10.1371/journal.pcbi.1010739.s011'
        resp = requests.get(url)

        if resp.status_code == 200:
            df = pd.read_excel(io.BytesIO(resp.content), sheet_name='Densities BBCAv1')
            pl.from_pandas(df).write_csv(output)
            fprint(f'Download successfully cellatlas csv and save in {output}!', vtype='io')
        else:
            raise RuntimeError('download cellatlas FAIL')

        return df

    @classmethod
    def load_sync_allen_structure_tree(cls, force_save: bool = True) -> pl.DataFrame:
        """
        TODO `ProS` not found
        Based on cellatlas dataframe, create a sync used `acronym` header in allen struct_tree (sorted by name)

        **fields**

        ``name``: `Brain region` found in cellatlas

        ``n_neurons``: `n_neurons` from cellatlas dataframe

        ``acronym``: `acronym` found in the structure_tree csv

        :param force_save: create sync file every time in root directory
        :return: sync_dataframe
            example::

                ┌────────────────────────────────┬────────────────┬───────────┬─────────┐
                │ name                           ┆ Volumes [mm^3] ┆ n_neurons ┆ acronym │
                │ ---                            ┆ ---            ┆ ---       ┆ ---     │
                │ str                            ┆ f64            ┆ i64       ┆ str     │
                ╞════════════════════════════════╪════════════════╪═══════════╪═════════╡
                │ Abducens nucleus               ┆ 0.015281       ┆ 1324      ┆ VI      │
                │ Agranular insular area         ┆ 4.901734       ┆ 242362    ┆ AI      │
                │ …                              ┆ …              ┆ …         ┆ …       │
                │ Visual areas                   ┆ 12.957203      ┆ 1297194   ┆ VIS     │
                │ Zona incerta                   ┆ 2.157641       ┆ 136765    ┆ ZI      │
                │ posteromedial visual area      ┆ 1.2225625      ┆ 197643    ┆ VISpm   │
                └────────────────────────────────┴────────────────┴───────────┴─────────┘
        """
        out = ATLAS_CACHE_DIRECTORY / 'cellatlas_allen_sync.csv'

        if not out.exists() or force_save:
            ctlas = cls.load_from_csv()

            allen = load_bg_structure_tree().select('name', 'acronym').sort('name')

            df = (
                ctlas.dataframe
                .rename({'Brain region': 'name'})
                .join(allen, on='name')
            )

            df.write_csv(out)
            fprint(f'SAVE cellatlas sync csv to {out}', vtype='io')

        return pl.read_csv(out)



from pathlib import Path
from typing import Literal, overload

import nrrd
import numpy as np
import polars as pl
from brainglobe_atlasapi import BrainGlobeAtlas

from neuralib.io import save_json, load_json
from neuralib.io.core import ALLEN_SDK_DIRECTORY, ATLAS_CACHE_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.deprecation import deprecated_func
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.verbose import fprint, print_save

__all__ = [
    'load_bg_structure_tree',
    'get_children',
    'get_annotation_ids',
    'get_leaf_in_annotation',
    'build_annotation_leaf_map',
    #
    'get_dorsal_cortex',
    #
    'load_allensdk_annotation',
    'load_ccf_annotation',
    'load_ccf_template',
    'load_structure_tree',

]


def load_bg_structure_tree(atlas_name: str = 'allen_mouse_10um',
                           check_latest: bool = True,
                           paired: bool = False) -> pl.DataFrame:
    """
    Load structure dataframe or dict from `brainglobe_atlasapi`

    :param atlas_name: allen source name
    :param check_latest: if check the brainglobe api latest version
    :param paired: To only ``acronym`` & ``parent_acronym`` fields
    :return:
    """
    file = BrainGlobeAtlas(atlas_name, check_latest=check_latest).root_dir / 'structures.csv'
    df = pl.read_csv(file).with_columns(pl.col('parent_structure_id').cast(pl.Int64))
    df = df.join(
        df.select([pl.col("id").alias("parent_structure_id"), pl.col("acronym").alias("parent_acronym")]),
        on="parent_structure_id",
        how="left"
    )

    if paired:
        name = df.select(pl.col('acronym'), pl.col('id'), pl.col('parent_structure_id'))
        join_df = name.join(name, left_on='parent_structure_id', right_on='id')
        parent_child = join_df.select(pl.col('acronym'), pl.col('names_right').alias('parent_acronym'))

        return parent_child
    else:
        return df


@overload
def get_children(parent: int, *, dataframe: bool = False, atlas_name: str = 'allen_mouse_10um') -> list[int] | pl.DataFrame:
    pass


@overload
def get_children(parent: str, *, dataframe: bool = False, atlas_name: str = 'allen_mouse_10um') -> list[str] | pl.DataFrame:
    pass


def get_children(parent: int | str, dataframe: bool = False, atlas_name: str = 'allen_mouse_10um') -> list[str] | pl.DataFrame:
    df = load_bg_structure_tree(atlas_name=atlas_name)
    return _get_children(df, parent, dataframe)


def _get_children(df, parent, dataframe):
    if isinstance(parent, int):
        ret = df.filter(pl.col('parent_structure_id') == parent)
        field = 'id'
    elif isinstance(parent, str):
        ret = df.filter(pl.col('parent_acronym') == parent)
        field = 'acronym'
    else:
        raise TypeError('')

    if not dataframe:
        ret = ret[field].to_list()

    return ret


# ============= #
# BG Annotation #
# ============= #

def get_annotation_ids(atlas_name: str = 'allen_mouse_10um', check_latest: bool = True) -> np.ndarray:
    annotation = BrainGlobeAtlas(atlas_name, check_latest=check_latest).annotation
    return np.unique(annotation)


def get_leaf_in_annotation(region: int | str, *,
                           name: bool = False,
                           cached_file: PathLike | None = None) -> list[int] | list[str]:
    """
    Get a list of annotation {id, acronym} with given region {id, acronym}

    :param region: region id or region acronym
    :param name: If True, return acronym, otherwise return id
    :param cached_file: cached json for the annotation_leaf_map
    :return: List of annotation {id, acronym}
    """
    tree = load_bg_structure_tree()

    # to id
    if isinstance(region, str):
        region_ids = tree.filter(pl.col('acronym') == region)['id'].to_list()
        if len(region_ids) != 1:
            raise RuntimeError(f"The region {region} is not a valid acronym")
        region = region_ids[0]

    dy = build_annotation_leaf_map(cached_file)

    try:
        result = dy[region]
    except KeyError:
        raise ValueError(f'Invalid region: {region}')

    if name:
        result = tree.filter(pl.col('id').is_in(result))['acronym'].to_list()

    return result


def build_annotation_leaf_map(cached_file: PathLike | None = None) -> dict[int, list[int]]:
    """
    Get all region id (key) and list of annotation id (values)

    :param cached_file: cached json file
    :return:
    """
    if cached_file is None:
        cached_file = ATLAS_CACHE_DIRECTORY / 'annotation_leaf.json'

    if Path(cached_file).suffix != '.json':
        raise ValueError('not a json file')

    #
    if cached_file.exists():
        data = load_json(cached_file, verbose=False)
        leaf_map = {int(k): v for k, v in data.items()}
    else:
        tree = load_bg_structure_tree()
        id_to_children = _build_id_to_children_map(tree)
        annotation_ids = set(get_annotation_ids())

        leaf_map = {}

        def collect(rid):
            if rid in leaf_map:
                return leaf_map[rid]
            if rid in annotation_ids:
                leaf_map[rid] = [rid]
            else:
                result = []
                for child in id_to_children.get(rid, []):
                    result.extend(collect(child))
                leaf_map[rid] = result
            return leaf_map[rid]

        all_ids = tree['id'].to_list()
        for rid in all_ids:
            collect(rid)

        save_json(cached_file, leaf_map)

    return leaf_map


def _build_id_to_children_map(tree: pl.DataFrame) -> dict[int, list[int]]:
    df = tree.select(['id', 'parent_structure_id'])
    grouped = df.group_by('parent_structure_id', maintain_order=False).agg(pl.col('id'))
    return {row['parent_structure_id']: row['id'] for row in grouped.iter_rows(named=True)}


# =============== #
# Allen Resources #
# =============== #

def get_dorsal_cortex(output_dir: Path | None = None) -> Path:
    """
    Get example dorsal projection annotation svg file

    .. seealso::

        https://community.brain-map.org/t/aligning-dorsal-projection-of-mouse-common-coordinate-framework-with-wide-field-images-of-mouse-brain/140/2

    :param output_dir: Output directory for caching
    :return: Output file path
    """

    if output_dir is None:
        output_dir = ATLAS_CACHE_DIRECTORY

    filename = 'cortical_map_top_down.svg'
    output = output_dir / filename

    if not output.exists():
        url = 'http://connectivity.brain-map.org/assets/cortical_map_top_down.svg'
        content = download_with_tqdm(url)

        with open(output, 'wb') as f:
            f.write(content.getvalue())
            print_save(output, verb='DOWNLOAD')

    return output


# ================ #
# TO BE DEPRECATED #
# ================ #

@deprecated_func(removal_version='0.4.3', remarks='switch brainglobe api instead')
def load_ccf_annotation(output_dir: PathLike | None = None) -> np.ndarray:
    from ._deprecate import _load_ccf_annotation
    return _load_ccf_annotation(output_dir)


@deprecated_func(removal_version='0.4.3', remarks='switch brainglobe api instead')
def load_ccf_template(output_dir: PathLike | None = None) -> np.ndarray:
    from ._deprecate import _load_ccf_template
    return _load_ccf_template(output_dir)


@deprecated_func(removal_version='0.4.3', remarks='switch brainglobe api instead')
def load_structure_tree(version: Literal['2017', 'old'] = '2017', output_dir: PathLike | None = None) -> pl.DataFrame:
    from ._deprecate import _load_structure_tree
    return _load_structure_tree(version, output_dir)


@deprecated_func(removal_version='0.4.3', remarks='switch brainglobe api instead, and probably deprecate allensdk dependency')
def load_allensdk_annotation(resolution: int = 10, output_dir: PathLike | None = None) -> np.ndarray:
    """
    Data Source directly from Allen Institute

    .. seealso::

        https://download.alleninstitute.org/informatics-archive/current-release/mouse_ccf/annotation/

    :param resolution: volume resolution in um. default is 10 um
    :param output_dir: output directory for caching
    :return: Array[uint32, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = ALLEN_SDK_DIRECTORY
        if not ALLEN_SDK_DIRECTORY.exists():
            ALLEN_SDK_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / f'annotation_{resolution}.nrrd'

    if not file.exists():
        try:
            from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
        except ImportError as e:
            fprint('Build error from project.toml. Please manually install using "pip install allensdk --no-deps"', vtype='error')
            raise e

        mcapi = MouseConnectivityApi()
        version = MouseConnectivityApi.CCF_VERSION_DEFAULT

        mcapi.download_annotation_volume(version, resolution, file)

    return nrrd.read(file)[0]



import polars as pl
from polars.testing import assert_frame_equal

from neuralib.util.verbose import printdf

__all__ = ['assert_polars_equal_verbose']


def assert_polars_equal_verbose(df1: pl.DataFrame, df2: pl.DataFrame):
    try:
        assert_frame_equal(df1, df2)
        print('DataFrames are equal.')
    except AssertionError as e:
        print('DataFrames are NOT equal.')

        # shape
        print('\nShape mismatch:')
        print(f'df1: {df1.shape}')
        print(f'df2: {df2.shape}')

        # column
        if df1.columns != df2.columns:
            print('\nColumn mismatch:')
            print(f'df1 columns: {df1.columns}')
            print(f'df2 columns: {df2.columns}')
            raise e

        df1_extra = df1.join(df2, on=df1.columns, how='anti')
        df2_extra = df2.join(df1, on=df1.columns, how='anti')

        if df1_extra.height > 0:
            print('\nRows in df1 not in df2:')
            printdf(df1_extra)

        if df2_extra.height > 0:
            print('\nRows in df2 not in df1:')
            printdf(df2_extra)

        # If shapes match, show cell-wise diff
        if df1.shape == df2.shape:
            print('\nCell-wise differences (non-equal values):')
            diffs = _highlight_cell_differences(df1, df2)
            print(diffs)

        raise e


def _highlight_cell_differences(df1: pl.DataFrame, df2: pl.DataFrame) -> pl.DataFrame:
    return pl.DataFrame({
        col: df1[col].cast(str).zip_with(
            df1[col] != df2[col],
            pl.lit('df1=') + df1[col].cast(str) + ', df2=' + df2[col].cast(str)
        ).fill_null('')  # Handle NaNs
        for col in df1.columns
    })



from typing import Literal

__all__ = ['Area', 'Source', 'Channel', 'MergeLevel',
           'HEMISPHERE_TYPE']

Area = str
"""Brain area"""

Source = str
"""Name of the injection alias"""

Channel = str
"""Fluorescence channel"""

MergeLevel = int
"""Hierarchical tree merge level"""

HEMISPHERE_TYPE = Literal['ipsi', 'contra', 'both']
"""brain hemisphere"""
